#include "dynexite_exam.hpp"

#include "util/enums.hpp"

#include <algorithm>
#include <ranges>
#include <stdexcept>
#include <string>

static void checkHeader(const std::string &found,
                        const std::string &expected) noexcept(false) {
    if (found != expected) {
        throw std::invalid_argument(
            "Read dynexite-CSV: Header validation error: expected \"" +
            expected + "\", found \"" + found + "\"");
    }
}

static void checkHeaderEndsWith(const std::string &found,
                                const std::string &ending) noexcept(false) {
    if (!found.ends_with(ending)) {
        throw std::invalid_argument(
            "Read dynexite-CSV: Header validation error: expected ending \"" +
            ending + "\", found string \"" + found + "\"");
    }
}

/**
 * \brief Check the naming of all headers.
 */
static void validateHeaders(const CsvDump &c,
                            const CsvLine::size_type nrTasks) noexcept(false) {

    // Do headers exist at all?
    if (c.data.empty())
        throw std::invalid_argument(
            "Read dynexite-CSV: Header validation: no headers found");

    const auto &headers = c.data[0];
    const auto nrColumns = c.data[0].size();

    checkHeader(headers[0], "AttemptID");
    checkHeader(headers[1], "Identifier");

    for (CsvLine::size_type i = 0; i < nrTasks; ++i) {
        checkHeaderEndsWith(headers[i * 4 + 2], " (open)");
        checkHeaderEndsWith(headers[i * 4 + 3], " (closed)");
        checkHeaderEndsWith(headers[i * 4 + 4], " (earned)");
        checkHeaderEndsWith(headers[i * 4 + 5], " (max)");
    }

    for (CsvLine::size_type i = 3;
         const auto &s : {"Open", "Closed", "Overall"}) {
        checkHeader(headers[nrColumns - 3 - i * 3], s);
        checkHeader(headers[nrColumns - 2 - i * 3], std::string(s) + " (max)");
        checkHeader(headers[nrColumns - 1 - i * 3], std::string(s) + " (%)");

        --i;
    }

    checkHeader(headers[nrColumns - 3], "Progress");
    checkHeader(headers[nrColumns - 2], "Execution Time");
    checkHeader(headers[nrColumns - 1], "Late Registration");
}

DynexiteExam::DynexiteExam(const CsvDump &c) noexcept(false) {
    if (c.data.empty())
        throw std::invalid_argument("Read dynexite-CSV: Empty file");

    const auto nrRows = c.data.size();
    const auto nrColumns = c.data[0].size();

    // No data aside from header?
    if (nrRows < 2)
        throw std::invalid_argument("Read dynexite-CSV: No data rows");

    // Non-attending students are listed as lines in the csv-file, but the lines
    // are empty aside from column 1, especially they are empty in line 0
    // (AttemptID)
    const auto nrStudents =
        std::ranges::count_if(c.data | std::views::drop(1),
                              [](const CsvLine &l) { return !l[0].empty(); });

    // No-one attended?
    if (nrStudents <= 0)
        throw std::invalid_argument("Read dynexite-CSV: No student attended?");

    // Assert that all necessary columns are there and the number of columns
    // makes sense (4 columns per task, 2 extra columns to the left and 12 extra
    // columns to the right)
    if (!(nrColumns > 14 && (nrColumns - 14) % 4 == 0))
        throw std::invalid_argument(
            "Read dynexite-CSV: Unexpected number of columns");

    const auto nrTasks = (nrColumns - 14) / 4;

    reserveArrays(static_cast<qsizetype>(nrStudents),
                  static_cast<qsizetype>(nrTasks));

    validateHeaders(c, nrTasks);

    if (!insertData(c, nrTasks)) {
        throw std::invalid_argument("Read dynexite-CSV: Inserting data failed");
    }

    /* todo: run some data sanity checks:
        - all task max points == total max points
        - all task earned points < task max points
        - all points are multiples of GradeBoundaryStepSize (no assert, just
       warn)
        - all maxPoints (of tasks or in total) are multiples of
       GradeBoundaryStepSize
        - etc */
}

/**
 * \brief Since we know the future sizes of all data structure, we reserve them
 * for some performance improvements. Also, we already create an empty (but
 * reserved) list of points for each task.
 */
void DynexiteExam::reserveArrays(const qsizetype nrStudents,
                                 const qsizetype nrTasks) noexcept {

    mAttemptIds.reserve(nrStudents);
    mIdentifier.reserve(nrStudents);

    using enum GradingType;

    mPoints[Closed].reserve(nrStudents);
    mPoints[Open].reserve(nrStudents);
    mPoints[Combined].reserve(nrStudents);

    mTaskMaxPoints.resize(nrTasks);

    // Note that we resize these instead of reserving, because they are lists of
    // lists.
    mTaskEarnedPoints[Closed].resize(nrTasks);
    mTaskEarnedPoints[Open].resize(nrTasks);
    mTaskEarnedPoints[Combined].resize(nrTasks);

    // For every list in the newly resized list of lists, we reserve the
    // required space.
    for (auto i = 0; i < nrTasks; ++i) {
        mTaskEarnedPoints[Closed][i].reserve(nrStudents);
        mTaskEarnedPoints[Open][i].reserve(nrStudents);
        mTaskEarnedPoints[Combined][i].reserve(nrStudents);
    }
}

/**
 * \brief Call std::stod, but interpret an empty string as 0.
 */
static Points stringToPoints(std::string s) noexcept(false) {

    if (s.empty())
        return Points{0.};
    // dynexite file might use dot or comma as decimal separator
    std::ranges::replace(s, ',', '.');
    return Points{std::stod(s)};
}

bool DynexiteExam::insertData(
    const CsvDump &c, const CsvLine::size_type nrTasks) noexcept(false) {

    const auto nrColumns = c.data[0].size();

    using enum GradingType;

    // Max points are stored in every line, but should be the same everywhere
    mPointsMax[Open] = stringToPoints(c.data[1][nrColumns - 11]);
    mPointsMax[Closed] = stringToPoints(c.data[1][nrColumns - 8]);
    mPointsMax[Combined] = stringToPoints(c.data[1][nrColumns - 5]);

    if (mPointsMax[Closed] + mPointsMax[Open] != mPointsMax[Combined])
        return false;

    // Ignore header line, ignore empty, non-attendant students
    for (const auto &line : c.data | std::views::drop(1)) {

        // Skip non-attending students
        if (line[0].empty())
            continue;

        mAttemptIds.push_back(QString::fromStdString(line[0]));
        mIdentifier.push_back(std::stoi(line[1]));

        // Points per task
        for (CsvLine::size_type taskIdxU = 0; taskIdxU < nrTasks; ++taskIdxU) {

            // The first two columns are not task related.
            static constexpr CsvLine::size_type startOffset = 2;
            const auto taskIdx = static_cast<qsizetype>(taskIdxU);

            /*
             * The columns are in the order:
             *      Earned open
             *      Earned closed
             *      Earned combined
             *      Max overall
             */
            mTaskEarnedPoints[Open][taskIdx].push_back(
                stringToPoints(line[startOffset + taskIdxU * 4]));
            mTaskEarnedPoints[Closed][taskIdx].push_back(
                stringToPoints(line[startOffset + taskIdxU * 4 + 1]));
            mTaskEarnedPoints[Combined][taskIdx].push_back(
                stringToPoints(line[startOffset + taskIdxU * 4 + 2]));

            // Max points are only shown for students who were assigned this
            // task, else the entry is empty, therefore only read the max points
            // of the task if the entry is not empty
            if (!line[startOffset + taskIdxU * 4 + 3].empty()) {
                if (mTaskMaxPoints[taskIdx] == Points{0.}) {
                    mTaskMaxPoints[taskIdx] =
                        stringToPoints(line[5 + taskIdxU * 4]);
                } else {
                    // Assert that all lines where the max points are shown show
                    // the same max points for a task
                    if (mTaskMaxPoints[taskIdx] !=
                        stringToPoints(line[startOffset + taskIdxU * 4 + 3]))
                        return false;
                }
            }
        }

        // Total exam points
        if (line[nrColumns - 12].empty()) {
            // An empty line indicates a non-attending student.
            // Lines with non-attending students should have been skipped!
            assert(false);

            // Anyways, if the student didn't attend, fill in zeros.
            mPoints[Closed].push_back(Points{0.});
            mPoints[Open].push_back(Points{0.});
            mPoints[Combined].push_back(Points{0.});

        } else {
            // Achieved open points
            mPoints[Open].push_back(stringToPoints(line[nrColumns - 12]));

            // Total open points should be the same everywhere
            if (mPointsMax[Open] != stringToPoints(line[nrColumns - 11]))
                return false;

            // Skip open points percentage at nrColumns - 10

            // Achieved closed points
            mPoints[Closed].push_back(stringToPoints(line[nrColumns - 9]));

            // Total closed points should be the same everywhere
            if (mPointsMax[Closed] != stringToPoints(line[nrColumns - 8]))
                return false;

            // Skip closed points percentage at nrColumns - 7

            // Achieved combined points
            mPoints[Combined].push_back(stringToPoints(line[nrColumns - 6]));

            // Total combined points should be the same everywhere
            if (mPointsMax[Combined] != stringToPoints(line[nrColumns - 5]))
                return false;

            // Skip overall points percentage at nrColumns - 4

            // Skip progress, execution time, late registration (unused)
        }
    }
    return true;
}
