#include "passing_limit.hpp"

#include "first_try.hpp"
#include "user_point_config.hpp"

#include "util/defaults.hpp"

#include <algorithm>
#include <numeric>

/*
 * Quotes from:
 * Handreichung zum Thema Multiple Choice Pruefungen im Rahmen des Handbuchs
 * fuer Studium und Lehre, Stichpunkt 5, version of 29th May 2019
 */

namespace impl {
// absolute passing limit (absolute Bestehensgrenze)
// 60 % of the maximum points
Points
calculateAbsolutePassingLimit(const GradeBoundary max,
                              const GradeBoundaryStepSize stepSize) noexcept {
    return toPoints(max, stepSize) * 0.6;
}

// relative passing limit (relative Bestehensgrenze)
//      "mindestens 50 % der maximal zu erreichenden Punktzahl erzielt
//      wurden und die erreichte Punktzahl um nicht mehr als 22 % die
//      durchschnittliche Punktzahl der Kandidatinnen und Kandidaten
//      unterschreitet, die erstmals an der Pruefung teilgenommen haben"
Points calculateRelativePassingLimit(
    const GradeBoundary max, const GradeBoundaryStepSize stepSize,
    const QList<Points> &firstTryClosedPoints) noexcept {

    // if there weren't any first-time exam writers, we don't have a separate
    // relative passing limit
    if (firstTryClosedPoints.empty())
        return calculateAbsolutePassingLimit(max, stepSize);

    const auto averageFirstTryPoints =
        std::reduce(firstTryClosedPoints.begin(), firstTryClosedPoints.end()) /
        firstTryClosedPoints.size();

    return std::max(toPoints(max, stepSize) * 0.5,
                    averageFirstTryPoints * 0.78);
}

// "Anwendung findet die fuer die Studierenden guenstigere
// Bestehensgrenze."
Points
calculatePassingLimit(const GradeBoundary max,
                      const GradeBoundaryStepSize stepSize,
                      const QList<Points> &firstTryClosedPoints) noexcept {
    return std::min(
        calculateRelativePassingLimit(max, stepSize, firstTryClosedPoints),
        calculateAbsolutePassingLimit(max, stepSize));
}
} // namespace impl

void PassingLimit<GradingType::Closed>::recalculate() {
    // Interestingly, we can only access the base class members using "this->"
    // here Read here for details:
    // https://isocpp.org/wiki/faq/templates#nondependent-name-lookup-members
    const auto &firstTryRef = std::get<const FirstTry &>(this->mParentRefs);
    const auto &maxRef =
        std::get<const ChangedMaxPoints<GradingType::Closed> &>(
            this->mParentRefs);
    const auto &configRef =
        std::get<const UserPointConfig &>(this->mParentRefs);

    // Closed passing limit: calculate
    if (configRef.mCountZerothAttemptAsFirstAttempt) {
        mRawPassingLimit = impl::calculatePassingLimit(
            maxRef.mChangedMaxPointBoundary, configRef.mGradeBoundaryStepSize,
            firstTryRef.mFirstTryClosedPoints +
                firstTryRef.mZerothTryClosedPoints);
    } else {
        mRawPassingLimit = impl::calculatePassingLimit(
            maxRef.mChangedMaxPointBoundary, configRef.mGradeBoundaryStepSize,
            firstTryRef.mFirstTryClosedPoints);
    }
    mPassingLimit =
        GradeBoundary{mRawPassingLimit, configRef.mGradeBoundaryStepSize,
                      configRef.mGradeBoundaryRounding};

    // mMaxPassingLimit is unused for closed points
    mMaxPassingLimit = mPassingLimit;
}

void PassingLimit<GradingType::Open>::recalculate() {
    const auto &maxRef = std::get<const ChangedMaxPoints<GradingType::Open> &>(
        this->mParentRefs);
    const auto &configRef =
        std::get<const UserPointConfig &>(this->mParentRefs);

    mRawPassingLimit = toPoints(maxRef.mChangedMaxPointBoundary,
                                configRef.mGradeBoundaryStepSize) *
                       defaults::openPointsStartingPassingLimit;

    mPassingLimit =
        GradeBoundary{mRawPassingLimit, configRef.mGradeBoundaryStepSize,
                      configRef.mGradeBoundaryRounding};

    const auto rawMaxPassingLimit = toPoints(maxRef.mChangedMaxPointBoundary,
                                             configRef.mGradeBoundaryStepSize) *
                                    defaults::openPointsPassingLimit;

    mMaxPassingLimit =
        GradeBoundary{rawMaxPassingLimit, configRef.mGradeBoundaryStepSize,
                      configRef.mGradeBoundaryRounding};
}
