#include "student_data.hpp"

#include <algorithm>
#include <array>
#include <ranges>
#include <stdexcept>
#include <string>
#include <utility>
#include <vector>

#define RWTHONLINE_PARSER_PERMISSIVE 1

[[maybe_unused]] static void checkHeader(const std::string &found,
                                         const std::string &expected) {
    if (found != expected) {
        throw std::invalid_argument(
            "Read rwthonline-CSV: Header validation error: expected \"" +
            expected + "\", found \"" + found + "\"");
    }
}

/**
 * \brief Check the naming of all headers.
 */
static std::pair<std::size_t, std::size_t> validateHeaders(CsvDump &c) {

    // Do headers exist at all?
    if (c.data.empty())
        throw std::invalid_argument(
            "Read rwthonline-CSV: Header validation: no headers found");

    auto &headers = c.data[0];

    // sometimes the header strings are in quotes, sometimes not
    for (auto &s : headers) {
        std::erase(s, '"');
    }

    // In permissive mode we only check for the names of headers we require and
    // ignore all other header names.
#if !RWTHONLINE_PARSER_PERMISSIVE
    if (c.data[0].size() != 24) {
        throw std::invalid_argument("Read rwthonline-CSV: Header validation: "
                                    "Unexpected number of columns: " +
                                    std::to_string(c.data[0].size()));
    }

    static constexpr std::array headerStrings = {"STUDY_PROGRAMME",
                                                 "CODE_OF_STUDY_PROGRAMME",
                                                 "Studienplan_Version",
                                                 "SPO_KONTEXT",
                                                 "REGISTRATION_NUMBER",
                                                 "FAMILY_NAME_OF_STUDENT",
                                                 "FIRST_NAME_OF_STUDENT",
                                                 "GESCHLECHT",
                                                 "DATE_OF_ASSESSMENT",
                                                 "GUEL_U_AKTUELLE_ANTRITTE_SPO",
                                                 "GRADE",
                                                 "REMARK",
                                                 "Number_Of_The_Course",
                                                 "SEMESTER_OF_The_COURSE",
                                                 "COURSE_TITLE",
                                                 "Examiner",
                                                 "Start_Time",
                                                 "TERMIN_ORT",
                                                 "DB_Primary_Key_Of_Exam",
                                                 "DB_Primary_Key_Of_Candidate",
                                                 "COURSE_GROUP_NAME",
                                                 "FILE_REMARK",
                                                 "EMAIL_ADDRESS",
                                                 "ECTS_GRADE"};

    for (int i = 0; i < 24; ++i) {
        checkHeader(headers[i], headerStrings[i]);
    }
    return {4u, 9u};
#else
    // find column of registration number and attempt nr
    const auto registrationNrIdx = static_cast<std::size_t>(
        std::ranges::find(c.data[0], "REGISTRATION_NUMBER") -
        std::ranges::begin(c.data[0]));

    const auto attemptNrIdx = static_cast<std::size_t>(
        std::ranges::find(c.data[0], "GUEL_U_AKTUELLE_ANTRITTE_SPO") -
        std::ranges::begin(c.data[0]));

    if (registrationNrIdx >= c.data[0].size()) {
        throw std::invalid_argument(
            "Read rwthonline-CSV: No registration number column found");
    }
    if (attemptNrIdx >= c.data[0].size()) {
        throw std::invalid_argument(
            "Read rwthonline-CSV: No attempt number column found");
    }
    return {registrationNrIdx, attemptNrIdx};
#endif
}

StudentData::StudentData(CsvDump &c) noexcept(false) {
    if (c.data.empty())
        throw std::invalid_argument("Read rwthonline-CSV: Empty file");

    const auto nrRows = c.data.size();

    // No data aside from header?
    if (nrRows < 2)
        throw std::invalid_argument("Read rwthonline-CSV: No data rows");

    const auto nrStudents = nrRows - 1;
    reserveArrays(static_cast<qsizetype>(nrStudents));

    const auto [regIdx, attemptIdx] = validateHeaders(c);

    insertData(c, regIdx, attemptIdx);
}

void StudentData::reserveArrays(const qsizetype nrStudents) noexcept {
    mRegistrationNr.reserve(nrStudents);
    mAttemptNr.reserve(nrStudents);
}

// std::stoi might throw
void StudentData::insertData(CsvDump &c, const std::size_t regIdx,
                             const std::size_t attemptIdx) noexcept(false) {

    // Read the registration number and attempt number from the csv-file to our
    // struct
    for (auto &line : c.data | std::views::drop(1)) {
        // The values may be enclosed in quotation marks in the csv-file
        std::erase(line[regIdx], '"');
        std::erase(line[attemptIdx], '"');

        mRegistrationNr.emplaceBack(std::stoi(line[regIdx]));
        mAttemptNr.emplaceBack(std::stoi(line[attemptIdx]));
    }
}
