/**
 * \file csv_parser.cpp
 * \brief Implements functions to read and store csv files from disk in memory,
 * including internal helper functions.
 */

#include "csv_parser.hpp"

#include <algorithm>
#include <cassert>
#include <cstddef>
#include <fstream>
#include <ranges>
#include <sstream>
#include <stdexcept>
#include <type_traits>
#include <utility>

/**
 * \brief To be able to read files with different line endings platform
 * independently, we use this helper function instead of std::getline
 *
 * \author https://stackoverflow.com/a/70582873
 */
std::istream &safeGetLine(std::istream &is, std::string &t) {
    t.clear();

    // The characters in the stream are read one-by-one using a std::streambuf.
    // That is faster than reading them one-by-one using the std::istream.
    // Code that uses streambuf this way must be guarded by a sentry object.
    // The sentry object performs various tasks,
    // such as thread synchronization and updating the stream state.

    std::istream::sentry se(is, true);
    std::streambuf *sb = is.rdbuf();

    for (;;) {
        int c = sb->sbumpc();
        switch (c) {
        case '\n':
            return is;
        case '\r':
            if (sb->sgetc() == '\n')
                sb->sbumpc();
            return is;
        case std::streambuf::traits_type::eof():
            is.setstate(std::ios::eofbit);
            if (t.empty())
                is.setstate(std::ios::failbit);
            return is;
        default:
            t += (char)c;
        }
    }
}

/**
 * \brief Splits the given string into a vector of strings using the given
 * delimiter.
 *
 * \param[in] text      String to split; use `std::move()` to avoid a copy.
 * \param[in] delim     Delimiter character.
 * \param[in] nrDelims  A guess at how many segments exist (optional, does not
 * have to be accurate). \param[in] allowEndWithDelim If the last character in
 * `text` is a delimiter, don't add or add an empty last string to the vector.
 *
 * \author Adopted from
 * https://stackoverflow.com/questions/5607589
 *
 * \return A vector of the string segments between the delimiter characters in
 * original order, with the delimiter characters removed.
 */
std::vector<std::string>
split(std::string text, const char delim, const bool allowEndWithDelim,
      const std::vector<std::string>::size_type nrDelims = 0) {
    assert(delim != '\n' && "For platform-independently splitting a string, "
                            "use safeGetLine instead of split!");

    // reserving enough space beforehand is optional, but improves performance
    std::vector<std::string> vec;
    vec.reserve(nrDelims);

    // pass by rvalue to avoid copying the string
    std::stringstream ss(std::move(text));

    // read until delimiter into line, then move the line to the end of the
    // vector
    std::string line;
    while (std::getline(ss, line, delim)) {
        vec.emplace_back(std::move(line));
    }

    // special case: if last character is the delimiter, we add an empty string
    // as last segment
    if (!allowEndWithDelim && ss.str().back() == delim) {
        vec.emplace_back();
    }

    return vec;
}

/**
 * \brief Reads every line of the file at the given path into a separate string
 * in a vector.
 *
 * \param[in] p Path to the file.
 *
 * \exception std::ios_base::failure Bad bit set while reading the file
 *
 * \exception std::invalid_argument The file could not be opened
 *
 * \return Each line of the file in a separate string, without the
 * newline-character.
 */
std::vector<std::string> readRaw(const std::filesystem::path &p) {
    // first, read whole file into a string for easier processing
    std::string fileStr;

    // try to open the file
    std::ifstream file(p);

    // throw invalid_argument if file couldn't be opened
    if (!file.is_open()) {
        throw std::invalid_argument("Failed to open file");
    }

    // throw std::ios_base::failure on future read errors
    file.exceptions(std::ifstream::badbit);

    // split by line
    std::string line;
    std::vector<std::string> vec;
    while (!safeGetLine(file, line).eof()) {
        vec.emplace_back(std::move(line));
    }
    // close file by leaving this scope
    return vec;
}

/**
 * \brief Counts the number of delimiters in each line, and throws an error if
 * the number is inconsistent.
 *
 * \todo Escaped delimiters.
 * \todo Maybe ignore empty lines.
 *
 * \param[in] fileLines One string per line of the original file
 * \param[in] delim     Delimiter character
 *
 * \exception std::runtime_error The number of delimiters per line is
 * inconsistent.
 *
 * \return Returns the consistent number of delimiters per line.
 */
std::vector<std::string>::size_type
getNrDelims(const std::vector<std::string> &fileLines, const char delim) {
    if (fileLines.empty())
        return 0;

    // take number of delimiters in the first line as baseline
    const auto nrDelims = std::ranges::count(fileLines[0], delim);

    // check that the number of delimiters in every other line is the same
    for (const auto &l : fileLines) {
        if (std::ranges::count(l, delim) != nrDelims) {
            // inconsistent number of delimiters (or escaped ones?)
            throw std::runtime_error(
                "Inconsistent number of delimiter characters");
        }
    }

    return static_cast<std::vector<std::string>::size_type>(nrDelims);
}

QSharedPointer<CsvDump> readCsv(const std::filesystem::path &p,
                                const char delim) {
    // store the resulting dump on the heap, as it might get quite big
    QSharedPointer<CsvDump> dump(new CsvDump());

    // read the given file line-by-line into a vector of strings, without caring
    // about the delimiter for now may throw std::ios_base::failure, which is
    // raised to the caller of this function
    auto fileLines = readRaw(p);

    const auto nrLines = fileLines.size();
    // obtain number of delimiters per line, or throw std::runtime_error if it
    // isn't consistent
    const auto nrDelims = getNrDelims(fileLines, delim);

    // write text without delimiters into the CsvDump
    dump->data.reserve(nrLines);
    for (auto &fileLine : fileLines) {
        dump->data.push_back(
            split(std::move(fileLine), delim, false, nrDelims));
        assert(dump->data.back().size() == static_cast<unsigned>(nrDelims + 1));
    }

    return dump;
}
