/**
 * \file input_tab_csv_single_cohort.cpp
 */

#include "input_tab_csv_single_cohort.hpp"

#include "util/enums.hpp"
#include "util/grade_boundary_step.hpp"
#include "util/heuristics.hpp"

#include <QLabel>

#include <algorithm>
#include <cassert>
#include <memory>

inline constexpr int dynexiteRow = 0;
inline constexpr int rwthOnlineRow = 1;
inline constexpr EnumStorage<int, GradeBoundaryStepSize> boundaryStepRow = {
    2, 3, 4};
inline constexpr int importRow = 5;

InputTabCsvSingleCohort::InputTabCsvSingleCohort(QWidget *parent)
    : InputTabBase(parent), mExamFileView(FileType::dynexite),
      mStudentFileView(FileType::rwth), mGradeBoundaryStepSize{
                                            QRadioButton{"1.0 point"},
                                            QRadioButton{"0.5 points"},
                                            QRadioButton{"0.25 points"}} {

    // Add widgets
    mMainLayout.addWidget(&mExamFileView, dynexiteRow, 0, 1, 2);
    mMainLayout.addWidget(&mStudentFileView, rwthOnlineRow, 0, 1, 2);

    mMainLayout.addWidget(new QLabel("Grade boundary step:"),
                          *std::ranges::min_element(boundaryStepRow.mData), 0,
                          3, 1);
    {
        using enum GradeBoundaryStepSize;

        for (const auto s : {Min1_00, Min0_50, Min0_25}) {

            mMainLayout.addWidget(&mGradeBoundaryStepSize[s],
                                  boundaryStepRow[s], 1, 1, 1);
            mGradeBoundaryStepSize[s].setCheckable(false);
        }
    }
    // Don't enable the import button as long as no files are selected.
    mImportButton.setEnabled(false);
    mMainLayout.addWidget(&mImportButton, importRow, 0, 1, 2);

    // Listen for signals
    connect(&mExamFileView, &LoadFileView::modelChange, this,
            &InputTabCsvSingleCohort::fileSelectChange);
    connect(&mStudentFileView, &LoadFileView::modelChange, this,
            &InputTabCsvSingleCohort::fileSelectChange);
    connect(&mImportButton, &QPushButton::clicked, this,
            &InputTabCsvSingleCohort::handleImportButton);

    // Temporary shortcut to speed up testing the app
    mExamFileView.tempSetDeveloperPaths();
    mStudentFileView.tempSetDeveloperPaths();
}

void InputTabCsvSingleCohort::handleImportButton() {

    // Get selected grade boundary step size
    auto gradeBoundaryStepSize = GradeBoundaryStepSize::Min0_25;
    {
        using enum GradeBoundaryStepSize;
        for (const auto s : {Min1_00, Min0_50, Min0_25}) {
            if (mGradeBoundaryStepSize[s].isChecked()) {
                gradeBoundaryStepSize = s;
                break;
            }
        }
    }

    // Careful: this calls the fileSelectChange-callback, which resets the radio
    // buttons. So make sure we read them out first.
    const auto dynexiteExam =
        std::get<std::unique_ptr<DynexiteExam>>(mExamFileView.get());
    const auto studentFile =
        std::get<std::unique_ptr<StudentData>>(mStudentFileView.get());

    if (dynexiteExam && studentFile)
        emit inputTabFilesImported(*dynexiteExam, *studentFile,
                                   gradeBoundaryStepSize);
    else {
        assert(false);
    }
}

void InputTabCsvSingleCohort::fileSelectChange() {

    // Update clickability of import button and grade boundary steps
    if (mExamFileView.mFileState == FileSelectState::ok &&
        mStudentFileView.mFileState == FileSelectState::ok) {

        // Enable the import button
        mImportButton.setEnabled(true);

        // Enable grade boundary buttons, select the radio button the heuristic
        // suggests
        for (auto &r : mGradeBoundaryStepSize) {
            r.setCheckable(true);
            r.setChecked(false);
        }
        const auto gradeBoundaryStepSize = guessGradeBoundaryStepSize(
            *std::get<DynexiteExam *>(mExamFileView.read()));
        mGradeBoundaryStepSize[gradeBoundaryStepSize].setChecked(true);

    } else {
        // Disable the import button
        mImportButton.setEnabled(false);

        // Disable the grade boundary buttons again.
        for (auto &r : mGradeBoundaryStepSize) {
            r.setChecked(false);
            r.setCheckable(false);
        }
    }
}
