#include "load_file_view.hpp"

#include "util/storage_helper.hpp"

#include <QFileDialog>
#include <QStyle>

#define HOME_MACHINE 1

QString getFileNameDesc(const FileType t) {
    switch (t) {
    case FileType::dynexite:
        return "Dynexite";
    case FileType::rwth:
        return "RWTHonline";
    case FileType::bgk:
        return "BGK";
    default:
        assert(false);
        return "Unknown";
    }
}

LoadFileView::LoadFileView(const FileType type, QWidget *parent)
    : QGroupBox(getFileNameDesc(type), parent), mInputModel(type) {

    // todo: set this, but it also does some annoying stuff - don't do them.
    // mFileNameLabel->setWordWrap(true);

    connect(&mBrowseButton, &QPushButton::clicked, this,
            &LoadFileView::handleBrowseButton);

    // the elements of each file selection are arranged horizontally within
    // the box
    mHLayout.addWidget(&mFileNameLabel);
    mHLayout.addWidget(&mFileParseStatusIcon);
    mHLayout.addWidget(&mBrowseButton);
    setLayout(&mHLayout);
}

void LoadFileView::handleBrowseButton() {
    // window captions for the file selection dialog for each file type
    static constexpr EnumStorage<const char *, FileType> captionTexts = {
        "Open exam results file from dynexite",
        "Open student info file from RWTHonline",
        "Open eTest bonus points file", "Open BGK file"};

    static constexpr auto csvFilter =
        "Comma-separated values (CSV) file (*.csv);;All files (*.*)";
    static constexpr auto bgkFilter =
        "Configured exam (BGK) file (*.bgk);;All files (*.*)";

    // let the user select a file
    const auto fileName = QFileDialog::getOpenFileName(
        this, captionTexts[mInputModel.mType], "",
        mInputModel.mType == FileType::bgk ? bgkFilter : csvFilter);

    // if the user didn't select any files, don't change anything
    if (!fileName.isEmpty()) {
        // write the path to the selected file to the label of the corresponding
        // file type
        mFileNameLabel.setText(fileName);

        if (mInputModel.loadCsv(mFileNameLabel.text().toStdString())) {
            setFileStatus(FileSelectState::ok);
        } else {
            setFileStatus(FileSelectState::error);
        }
    }
}

void LoadFileView::reset() {
    // remove input from the labels displaying the selected files, reset icons,
    // unload csv-file
    mFileNameLabel.setText("");

    // We deliberately keep the current data loaded in the model, so that we can
    // reset the view and extract the data from the model (which clears it
    // there)

    setFileStatus(FileSelectState::empty);
}

void LoadFileView::tempSetDeveloperPaths() {
#if defined(DYNEXITE_EXAMPLE_FILE_PATH) && defined(RWTHONLINE_EXAMPLE_FILE_PATH)
    if (mInputModel.mType == FileType::dynexite) {
        mFileNameLabel.setText(DYNEXITE_EXAMPLE_FILE_PATH);
    } else {
        mFileNameLabel.setText(RWTHONLINE_EXAMPLE_FILE_PATH);
    }

    if (mInputModel.loadCsv(mFileNameLabel.text().toStdString())) {
        setFileStatus(FileSelectState::ok);
    } else {
        setFileStatus(FileSelectState::error);
    }
#endif
}

void LoadFileView::setFileStatus(const FileSelectState status) {
    // icons corresponding to each file select state
    static constexpr EnumStorage<QStyle::StandardPixmap, FileSelectState>
        icons = {
            QStyle::SP_MessageBoxQuestion, // empty, no icon, not used
            QStyle::SP_MessageBoxCritical,
            QStyle::SP_DialogOkButton // todo: find better icon
        };

    // update status
    mFileState = status;

    // update icon
    if (status != FileSelectState::empty) {
        mFileParseStatusIcon.setPixmap(
            style()->standardIcon(icons[status]).pixmap(16));
    } else {
        mFileParseStatusIcon.setText(" ");
    }
    emit modelChange();
}
