/**
 * \file main_window.cpp
 * \brief Definition of the \ref MainWindow member functions.
 */

#include "main_window.hpp"

#include "data/data_manager.hpp"
#include "util/grade.hpp"
#include "util/grade_boundary.hpp"
#include "util/heuristics.hpp"

#include <QAction>
#include <QClipboard>
#include <QGuiApplication>
#include <QMenuBar>

#include <algorithm>
#include <cmath>

// Surround the given string evenly with spaces, until it has a size of `size`
QString padToSize(QString str, const int size) {
    const auto spacesLeft = static_cast<int>(
        std::max(0., std::floor(static_cast<double>(size - str.size()) / 2.)));
    const auto spacesRight = static_cast<int>(
        std::max(0., std::ceil(static_cast<double>(size - str.size()) / 2.)));

    str.prepend(QString(spacesLeft, ' '));
    str.append(QString(spacesRight, ' '));

    return str;
}

void printBoundaries(QString &str, const GradingType type, const bool hasType,
                     const GradeBoundaryStepSize stepSize,
                     const int decimalPlaces) {

    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    if (hasType) {
        str += QString::fromStdString(toString(type)) +
               ":\n"
               "Grade --- Min. points --- Max. points\n";

        auto lastGradeBoundary =
            cur->mChangedMaxPointsOpen.mChangedMaxPointBoundary;

        for (const auto g : grades) {

            const auto minBoundary =
                toPoints(cur->getUserGradeBoundaries(type)[g], stepSize).p;

            // For max points of 1.0, we use exam max points, not changed max
            // points.
            // For other max points, we go one grade boundary step below the
            // next higher grade.
            const auto maxBoundary =
                g == Grade::grade1_0 ? cur->mExam.mPointsMax[type].p
                                     : toPoints(lastGradeBoundary, stepSize).p -
                                           as<double>(stepSize);

            str +=
                padToSize(toString(g), 7) + ":" +
                padToSize(QLocale().toString(minBoundary, 'f', decimalPlaces),
                          15) +
                "-" +
                padToSize(QLocale().toString(maxBoundary, 'f', decimalPlaces),
                          13) +
                "\n";
            lastGradeBoundary = cur->getUserGradeBoundaries(type)[g];
        }

        str += "\n";
    }
}

void copyGradingSchemeToClipboard() {
    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    auto *clipboard = QGuiApplication::clipboard();

    const bool hasClosed =
        cur->mExam.mPointsMax[GradingType::Closed] > Points{0.};
    const bool hasOpen = cur->mExam.mPointsMax[GradingType::Open] > Points{0.};

    QString gradingSchemeText = "";

    const auto stepSize = cur->mUserPointConfig.mGradeBoundaryStepSize;
    const auto decimalPlaces = nrDecimalPlaces(stepSize);

    printBoundaries(gradingSchemeText, GradingType::Open, hasOpen, stepSize,
                    decimalPlaces);
    printBoundaries(gradingSchemeText, GradingType::Closed, hasClosed, stepSize,
                    decimalPlaces);

    if (!hasOpen && !hasClosed) {
        gradingSchemeText +=
            "Selected exam has neither open nor closed points.";
    }

    clipboard->setText(gradingSchemeText);
}

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent), mInputWindow(*this, parent), mBoundaryWindow(parent),
      mViewSettings(parent), mGradeTab(mViewSettings) {

    setWindowTitle("Dynexite2RWTHonline");

    setCentralWidget(&mGradeTab);
    // handleCommandLineArgs();

    mBoundaryWindow.setWindowFlags(
        Qt::WindowStaysOnTopHint | Qt::WindowCloseButtonHint |
        Qt::WindowMinimizeButtonHint | Qt::MSWindowsFixedSizeDialogHint);
    mBoundaryWindow.setWindowFlags(mBoundaryWindow.windowFlags() &
                                   ~Qt::WindowMaximizeButtonHint);

    mInputWindow.setWindowFlags(
        Qt::WindowStaysOnTopHint | Qt::WindowCloseButtonHint |
        Qt::WindowMinimizeButtonHint | Qt::MSWindowsFixedSizeDialogHint);
    mInputWindow.setWindowFlags(mInputWindow.windowFlags() &
                                ~Qt::WindowMaximizeButtonHint);

    mViewSettings.setWindowFlags(
        Qt::WindowStaysOnTopHint | Qt::WindowCloseButtonHint |
        Qt::WindowMinimizeButtonHint | Qt::MSWindowsFixedSizeDialogHint);
    mViewSettings.setWindowFlags(mViewSettings.windowFlags() &
                                 ~Qt::WindowMaximizeButtonHint);

    // start input file selection
    mInputWindow.show();

    // Menu bar
    // "New": loads a new exam
    const auto newAct = new QAction("&New", this);
    newAct->setShortcuts(QKeySequence::New);
    connect(newAct, &QAction::triggered, this, &MainWindow::menuNewFile);

    // "Options": change the required points for e.g. 2.7, change requirePassAll
    const auto boundaryAct = new QAction("Options", this);
    connect(boundaryAct, &QAction::triggered, this,
            &MainWindow::menuOpenBoundaries);

    // "View settings": change visual options
    const auto viewSettingsAct = new QAction("Settings", this);
    connect(viewSettingsAct, &QAction::triggered, this,
            &MainWindow::menuOpenViewSettings);

    // "Export"
    const auto exportGradingSchemeClipboardAct =
        new QAction("To Clipboard", this);
    connect(exportGradingSchemeClipboardAct, &QAction::triggered,
            copyGradingSchemeToClipboard);

    mFileMenu = menuBar()->addMenu("&Exam");
    mFileMenu->addAction(newAct);
    mFileMenu->addAction(boundaryAct);
    mViewMenu = menuBar()->addMenu("&View");
    mViewMenu->addAction(viewSettingsAct);
    mExportMenu = menuBar()->addMenu("&Export");
    mExportGradingMenu = mExportMenu->addMenu("Grading scheme");
    mExportGradingMenu->addAction(exportGradingSchemeClipboardAct);
}

void MainWindow::resetToInputFiles(
    const DynexiteExam &exam, const StudentData &students,
    const GradeBoundaryStepSize gradeBoundaryStepSize) {

    UserPointConfig pc;
    pc.mGradeBoundaryStepSize = gradeBoundaryStepSize;
    if (DataManager::loadNewExam(exam, students, UserGradeConfig{}, pc)) {
        DataManager::makeNewCurrent();

        // close input window and show main window
        mInputWindow.hide();
        showMaximized();
    }
}

void MainWindow::menuNewFile() { mInputWindow.show(); }

void MainWindow::menuOpenBoundaries() { mBoundaryWindow.show(); }

void MainWindow::menuOpenViewSettings() { mViewSettings.show(); }

void MainWindow::closeEvent(QCloseEvent *event) {
    mInputWindow.hide();
    mBoundaryWindow.hide();
    mViewSettings.hide();

    QMainWindow::closeEvent(event);
}

void MainWindow::resizeEvent(QResizeEvent *event) {
    QMainWindow::resizeEvent(event);

    mGradeTab.handleResize();
}

/* void MainWindow::handleCommandLineArgs() const {
    // get list of command line arguments
    QStringList fileNames = QCoreApplication::arguments();

    // skip first argument (path of executable)
    if (fileNames.size() > 1) {
        mBgkFileChooserTab->newInputFile(fileNames[1]);
    }
}*/
