#include "testable_objects.hpp"

#include "data/graph.hpp"
#include "file_io/csv_parser.hpp"
#include "util/heuristics.hpp"

#include "gtest/gtest.h"

#include <filesystem>
#include <iostream>

auto getDynexiteCsvs() {
    std::vector<std::pair<std::string, QSharedPointer<const CsvDump>>>
        dynexiteDumps;

    for (const auto &dynexiteEntry : std::filesystem::directory_iterator(
             std::filesystem::path(TESTING_DYNEXITE_DIRECTORY))) {
        if (dynexiteEntry.path().extension() == "") {
            continue;
        }
        if (dynexiteEntry.path().extension() != ".csv") {
            std::cerr << "Unknown file extension: "
                      << dynexiteEntry.path().extension() << std::endl;
            continue;
        }

        // if readCsv fails, just skip the tests (tested separately elsewhere)
        try {
            auto dynexiteDump = readCsv(dynexiteEntry);
            EXPECT_TRUE(!!dynexiteDump)
                << "dynexite readCsv created a NULL-pointer";

            dynexiteDumps.emplace_back(dynexiteEntry.path().stem().string(),
                                       dynexiteDump);

        } catch (const std::exception &e) {
            std::cerr << "Error during readCsv for the file: "
                      << dynexiteEntry.path() << ": " << e.what() << std::endl;
            continue;
        }
    }

    return dynexiteDumps;
}
auto getRwthOnlineCsvs() {
    std::vector<std::pair<std::string, QSharedPointer<const CsvDump>>>
        rwthDumps;

    for (const auto &rwthEntry : std::filesystem::directory_iterator(
             std::filesystem::path(TESTING_RWTHONLINE_DIRECTORY))) {
        if (rwthEntry.path().extension() == "") {
            continue;
        }
        if (rwthEntry.path().extension() != ".csv") {
            std::cerr << "Unknown file extension: "
                      << rwthEntry.path().extension() << std::endl;
            continue;
        }

        // if readCsv fails, just skip the tests (tested separately elsewhere)
        try {
            auto rwthDump = readCsv(rwthEntry);
            EXPECT_TRUE(!!rwthDump)
                << "rwthonline readCsv created a NULL-pointer";

            rwthDumps.emplace_back(rwthEntry.path().stem().string(), rwthDump);

        } catch (const std::exception &e) {
            std::cerr << "Error during readCsv for the file: "
                      << rwthEntry.path() << ": " << e.what() << std::endl;
            continue;
        }
    }

    return rwthDumps;
}

// load exams only once for better performance
// load all exams found in the dynexite_data directory
auto getExams() {
    std::vector<std::pair<std::string, DAG>> configuredExams;

    if (dynexiteCsvs().size() != rwthOnlineCsvs().size()) {
        ADD_FAILURE() << "Different number of dynexite and rwthonline csvs "
                         "found! Aborting tests.";
        return configuredExams;
    }

    for (std::size_t i = 0; i < dynexiteCsvs().size(); ++i) {

        if (dynexiteCsvs()[i].first != rwthOnlineCsvs()[i].first) {
            ADD_FAILURE()
                << "Name mismatch of dynexite csv and rwthonline csv file! "
                   "Dynexite file name: "
                << dynexiteCsvs()[i].first
                << ", rwthOnline file name: " << rwthOnlineCsvs()[i].first;
            continue;
        }

        // if the ctor fails, just skip the tests (tested separately elsewhere)
        try {
            auto nonConstRwthOnlineCsvs = *rwthOnlineCsvs()[i].second;
            auto dynexiteExam = DynexiteExam(*dynexiteCsvs()[i].second);
            UserGradeConfig gc;
            UserPointConfig pc;
            pc.mGradeBoundaryStepSize =
                guessGradeBoundaryStepSize(dynexiteExam);

            auto dag =
                DAG{dynexiteExam, StudentData(nonConstRwthOnlineCsvs), gc, pc};

            configuredExams.emplace_back(dynexiteCsvs()[i].first, dag);

        } catch (const std::exception &e) {
            std::cerr << "Error during DAG construction for the file: "
                      << dynexiteCsvs()[i].first << ": " << e.what()
                      << std::endl;
            continue;
        }
    }

    // print file names
    std::cout << "Nr valid exam files found: " << configuredExams.size()
              << std::endl;
    for (auto &configuredExam : configuredExams) {
        std::cout << configuredExam.first << std::endl;
    }

    return configuredExams;
}

// vector of all found valid ConfiguredExams together with their file names
const std::vector<std::pair<std::string, QSharedPointer<const CsvDump>>> &
dynexiteCsvs() {
    static auto ret = getDynexiteCsvs();
    return ret;
}
const std::vector<std::pair<std::string, QSharedPointer<const CsvDump>>> &
rwthOnlineCsvs() {
    static auto ret = getRwthOnlineCsvs();
    return ret;
}
const std::vector<std::pair<std::string, DAG>> &exams() {
    static auto ret = getExams();
    return ret;
}
